package TeamDoodd_2021.extaction;

import TeamDoodd_2021.TeamDooddConstants;
import TeamDoodd_2021.debugger.DebugHelper;
import TeamDoodd_2021.extaction.pf.guidelineHelper;
import TeamDoodd_2021.module.algorithm.fb.TeamDooddCompositeConvexHull;
import TeamDoodd_2021.module.complex.fb.tools.FbUtilities;
import TeamDoodd_2021.module.complex.pf.TeamDooddGuidelineCreator;
import TeamDoodd_2021.standard.Ruler;
import TeamDoodd_2021.util.Util;
import TeamDoodd_2021.world.CSUWorldHelper;
import TeamDoodd_2021.world.object.CSUEdge;
import TeamDoodd_2021.world.object.CSULineOfSightPerception;
import TeamDoodd_2021.world.object.CSURoad;
import TeamDoodd_2021.world.CSUWorldHelper;
import adf.Main;
import adf.agent.action.Action;
import adf.agent.action.common.ActionMove;
import adf.agent.info.WorldInfo;
import com.mrl.debugger.remote.dto.StuckDto;
import rescuecore2.misc.Pair;
import rescuecore2.misc.geometry.GeometryTools2D;
import rescuecore2.misc.geometry.Line2D;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.misc.geometry.Vector2D;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;

import java.awt.*;
import java.util.List;
import java.util.*;
import java.util.stream.Collectors;

/**
 * @author: Guanyu-Cai
 * @Date: 03/10/2020
 */
public class StuckHelper {
    private TeamDooddGuidelineCreator TeamDooddGuidelineCreator;
    private CSUWorldHelper world;
    private CSULineOfSightPerception lineOfSightPerception;
    private WorldInfo worldInfo;
    public StuckHelper(CSUWorldHelper world, WorldInfo worldInfo) {
        this.world = world;
        this.worldInfo = worldInfo;
        this.lineOfSightPerception = new CSULineOfSightPerception(world);
    }
    private Point2D getMidPoint(Edge edge) {
        if(edge != null) {
            double midX = (edge.getStartX() + edge.getEndX()) / 2;
            double midY = (edge.getStartY() + edge.getEndY()) / 2;
            Point2D point = new Point2D(midX, midY);
            return point;
        }
        return null;
    }
    private Edge getOppositeEdge(Road road,Edge original) {
        List<Edge> edges = new ArrayList<>();
        for(Edge edge : road.getEdges()){
            if(edge.isPassable() && !this.getMidPoint(edge).equals(this.getMidPoint(original))){
                edges.add(edge);
            }
        }
        if(edges.size() > 0) {
            Point2D roadCenter = new Point2D(road.getX(), road.getY());
            Point2D originalMid = this.getMidPoint(original);
            Vector2D standardDirection = new Vector2D(roadCenter.getX() - originalMid.getX(), roadCenter.getY() - originalMid.getY());
            Edge answerEdge = null;
            double minAngle = Double.MAX_VALUE;
            for (int i = 0; i < edges.size(); ++i) {
                Point2D mid = this.getMidPoint(edges.get(i));
                Vector2D testDirection = new Vector2D(mid.getX() - originalMid.getX(), mid.getY() - originalMid.getY());
                double angle = GeometryTools2D.getAngleBetweenVectors(standardDirection, testDirection);
                if (angle < minAngle) {
                    minAngle = angle;
                    answerEdge = edges.get(i);
                }
            }
            return answerEdge;
        }
        return null;
    }
    private double getDistance(double fromX, double fromY, double toX, double toY) {
        double dx = fromX - toX;
        double dy = fromY - toY;
        return Math.hypot(dx, dy);
    }

    private Line2D getProperLine(Road road) {
        List<Edge> edges = new ArrayList<>();
        List<Edge> reservedEdges = new ArrayList<>();
        for (Edge edge : road.getEdges()) {
            Boolean edgeToEntrance = false;
            for(EntityID neighbour : road.getNeighbours()){
                Edge neighbourEdge = road.getEdgeTo(neighbour);
                if(!(this.worldInfo.getEntity(neighbour) instanceof Road)&& this.getMidPoint(edge).equals(this.getMidPoint(neighbourEdge))){
                    edgeToEntrance = true;
                    break;
                }
            }
            if(edge.isPassable()){
                reservedEdges.add(edge);
                if(!edgeToEntrance){
                    edges.add(edge);
                }
            }
        }

        Point2D start = null;
        Point2D end = null;
        double max = Double.MIN_VALUE;
        if(edges != null && edges.size() > 1) {
            for (Edge edge : edges) {
                Edge opposite = this.getOppositeEdge(road, edge);
                if(opposite != null) {
                    Point2D p1 = this.getMidPoint(edge);
                    Point2D p2 = this.getMidPoint(opposite);
                    double dist = this.getDistance(p1.getX(), p1.getY(), p2.getX(), p2.getY());
                    if (dist > max) {
                        max = dist;
                        start = p1;
                        end = p2;
                    }
                }
            }
            if (start != null && end != null) {
                Line2D guideline = new Line2D(start, end);
                return guideline;
            }
        }else if(edges != null && edges.size() == 1){
            Edge only = edges.get(0);
            Point2D onlyMid = this.getMidPoint(only);
            Point2D roadCenter = new Point2D(road.getX(),road.getY());
            List<Edge> allOtherEdges = new ArrayList<>();
            for(Edge e : road.getEdges()){
                if(!this.getMidPoint(e).equals(this.getMidPoint(only))){
                    allOtherEdges.add(e);
                }
            }
            Vector2D standardDirection = new Vector2D(roadCenter.getX() - onlyMid.getX(), roadCenter.getY() - onlyMid.getY());
            Point2D answerPoint = null;
            double minAngle = Double.MAX_VALUE;
            for (int i = 0; i < allOtherEdges.size(); ++i) {
                Point2D mid = this.getMidPoint(allOtherEdges.get(i));
                Vector2D testDirection = new Vector2D(mid.getX() - onlyMid.getX(), mid.getY() - onlyMid.getY());
                double angle = GeometryTools2D.getAngleBetweenVectors(standardDirection, testDirection);
                if (angle < minAngle) {
                    minAngle = angle;
                    answerPoint = mid;
                }
            }
            Line2D guideline = new Line2D(onlyMid,answerPoint);
            return guideline;
        }
        //in case of all the neighbours are entrance
        else {
            if(reservedEdges != null && reservedEdges.size() > 1) {
                for (Edge edge : reservedEdges) {
                    Edge opposite = this.getOppositeEdge(road, edge);
                    Point2D p1 = this.getMidPoint(edge);
                    Point2D p2 = this.getMidPoint(opposite);
                    double dist = this.getDistance(p1.getX(), p1.getY(), p2.getX(), p2.getY());
                    if (dist > max) {
                        max = dist;
                        start = p1;
                        end = p2;
                    }
                }
                if (start != null && end != null) {
                    Line2D guideline = new Line2D(start, end);
                    return guideline;
                }
            }else if(reservedEdges != null && reservedEdges.size() == 1){
                Edge only = reservedEdges.get(0);
                Point2D onlyMid = this.getMidPoint(only);
                Point2D roadCenter = new Point2D(road.getX(),road.getY());
                List<Edge> allOtherEdges = new ArrayList<>();
                for(Edge e : road.getEdges()){
                    if(!this.getMidPoint(e).equals(this.getMidPoint(only))){
                        allOtherEdges.add(e);
                    }
                }
                allOtherEdges.remove(only);
                Vector2D standardDirection = new Vector2D(roadCenter.getX() - onlyMid.getX(), roadCenter.getY() - onlyMid.getY());
                Point2D answerPoint = null;
                double minAngle = Double.MAX_VALUE;
                for (int i = 0; i < allOtherEdges.size(); ++i) {
                    Point2D mid = this.getMidPoint(allOtherEdges.get(i));
                    Vector2D testDirection = new Vector2D(mid.getX() - onlyMid.getX(), mid.getY() - onlyMid.getY());
                    double angle = GeometryTools2D.getAngleBetweenVectors(standardDirection, testDirection);
                    if (angle < minAngle) {
                        minAngle = angle;
                        answerPoint = mid;
                    }
                }
                Line2D guideline = new Line2D(onlyMid,roadCenter);
                return guideline;
            }
        }
        return null;
    }
    public Action calc(List<EntityID> path) {
        if (path == null || path.size() < 2) {
            return null;
        }

        List<EntityID> selfAndNeighborBlockades = getSelfAndNeighborBlockades();
        EntityID nearestBlockade = getNearestBlockade();
        List<StandardEntity> nearBlockades = getBlockadesInRange(nearestBlockade, selfAndNeighborBlockades, TeamDooddConstants.AGENT_PASSING_THRESHOLD);
        if (nearBlockades.isEmpty()) {
            return null;
        }
        TeamDooddCompositeConvexHull blockadesConvexHull = getConvexByBlockades(nearBlockades);

        StandardEntity selfPosition = world.getSelfPosition();
        Pair<Integer, Integer> selfLocation = world.getSelfLocation();
        Point2D locationPoint = new Point2D(selfLocation.first(), selfLocation.second());
        Point2D openPartCenter = null;
        Point2D edgeStart = null;
        Point2D edgeEnd = null;
        CSUEdge targetEdge = null;
        if (selfPosition instanceof Road) {
            CSURoad csuRoad = world.getCsuRoad(selfPosition);
            for (CSUEdge csuEdge : csuRoad.getCsuEdgesTo(path.get(1))) {
                if (!csuEdge.isBlocked()) {
                    openPartCenter = csuEdge.getOpenPartCenter();
                    edgeStart = csuEdge.getStart();
                    edgeEnd = csuEdge.getEnd();
                    targetEdge = csuEdge;
                    break;
                }
            }
        }
        if(selfPosition instanceof Building){
            Building sd = (Building) selfPosition;
            for(EntityID ne : sd.getNeighbours()){
                StandardEntity neigbor = world.getEntity(ne);
                if(neigbor instanceof Road){
                    CSURoad csuRoad = world.getCsuRoad(neigbor);
                    for (CSUEdge csuEdge : csuRoad.getCsuEdgesTo(sd.getID())) {
                        if (!csuEdge.isBlocked()) {
                            openPartCenter = csuEdge.getOpenPartCenter();
                            edgeStart = csuEdge.getStart();
                            edgeEnd = csuEdge.getEnd();
                            targetEdge = csuEdge;
                            break;
                        }
                    }
                }
            }
        }
        if (openPartCenter == null) {
            return null;
        }
        double x_dis=0,y_dist=0;
        if(edgeStart!=null)
            x_dis = (edgeEnd.getX()-edgeStart.getX())/8;
            y_dist = (edgeEnd.getY()-edgeStart.getY())/8;
        Line2D guideLine = new Line2D(locationPoint, openPartCenter);
        Point2D target = null;
        Set<Set<CSULineOfSightPerception.CsuRay>> raysNotHits = new HashSet<>();
        Set<CSULineOfSightPerception.CsuRay> raysNotHit1 = new HashSet<>();
        Collection<CSURoad> targetValidRoads = new HashSet<>();
        if (!Util.hasIntersectLine(blockadesConvexHull.getConvexPolygon(), guideLine)) {//由于使用了凸包,会漏判一些实际能走的情况
            target = Util.clipLine(guideLine, world.getConfig().maxRayDistance).getEndPoint();
            if (target.getX() > world.getMaxX() || target.getX() < 0 || target.getY() > world.getMaxY() || target.getY() < 0) {
                target = Util.improveLine(guideLine, TeamDooddConstants.AGENT_SIZE).getEndPoint();
            }
        } else {
            CSURoad csuRoad,oppositePassableEdgeRoad = null;
            if(selfPosition instanceof Road) {
                csuRoad = world.getCsuRoad(selfPosition);
                targetValidRoads.add(csuRoad);
                oppositePassableEdgeRoad = csuRoad.getOppositePassableEdgeRoad(targetEdge);
            }
            if(selfPosition instanceof Building){
                Building sd = (Building) selfPosition;
                for(EntityID ne : sd.getNeighbours()){
                    StandardEntity neigbor = world.getEntity(ne);
                    if(neigbor instanceof Road){
                        csuRoad = world.getCsuRoad(neigbor);
                        for (CSUEdge csuEdge : csuRoad.getCsuEdgesTo(sd.getID())) {
                            if (!csuEdge.isBlocked()) {
                                targetValidRoads.add(csuRoad);
                                break;
                            }
                        }
                    }
                }
            }
            if (oppositePassableEdgeRoad != null) {
                targetValidRoads.add(oppositePassableEdgeRoad);
            }
            int distance = 30000000;
            Line2D guidline = null;
            double x_dis_2 = 0,y_dis_2 = 0;
            raysNotHit1 = lineOfSightPerception.findRaysNotHit(locationPoint, nearBlockades, distance);
            Set<CSULineOfSightPerception.CsuRay> raysNotHit2 = lineOfSightPerception.findRaysNotHit(openPartCenter, nearBlockades, distance);
            for(int i = 0;i <= 8;i++) {
                Point2D now_point = new Point2D(edgeStart.getX()+x_dis*i,edgeStart.getY()+y_dist*i);
                Set<CSULineOfSightPerception.CsuRay> raysNotHit3 = lineOfSightPerception.findRaysNotHit(now_point, nearBlockades, distance);
                if(guidline != null)
                {
                    Point2D now_line = new Point2D(guidline.getOrigin().getX()+x_dis_2*i,guidline.getOrigin().getY()+y_dis_2*i);
                    Set<CSULineOfSightPerception.CsuRay> raysNotHit4 = lineOfSightPerception.findRaysNotHit(now_line, nearBlockades, distance);
                    raysNotHits.add(raysNotHit4);
                }
                raysNotHits.add(raysNotHit3);
            }
            raysNotHits.add(raysNotHit2);
            for (Set<CSULineOfSightPerception.CsuRay> next : raysNotHits) {
                List<Point2D> validIntersections = getValidIntersections(raysNotHit1, next, targetValidRoads);
                if (!validIntersections.isEmpty()) {
                    validIntersections.sort(new DistanceComparator(locationPoint));
                    target = Util.improveLine(new Line2D(locationPoint, validIntersections.get(0)), TeamDooddConstants.AGENT_PASSING_THRESHOLD).getEndPoint();
                }
            }
        }
        if (DebugHelper.DEBUG_MODE) {
            StuckDto stuckDto = new StuckDto();
            if (target != null) {
                stuckDto.setAgentId(world.getSelfHuman().getID().getValue());
                stuckDto.setBlockadesConvexHull(blockadesConvexHull.getConvexPolygon());
                stuckDto.setTarget(Util.convertPoint(target));
                stuckDto.setGuideLine(Util.convertLine2(guideLine));
                HashSet<java.awt.geom.Line2D> line2DS = new HashSet<>();
                for (Set<CSULineOfSightPerception.CsuRay> e : raysNotHits) {
                    line2DS.addAll(e.stream().map(a -> new java.awt.geom.Line2D.Double(a.getRay().getOrigin().getX(), a.getRay().getOrigin().getY(),
                            a.getRay().getEndPoint().getX(), a.getRay().getEndPoint().getY())).collect(Collectors.toSet()));
                }
                HashSet<java.awt.geom.Line2D> selfLine2DS = new HashSet<>();
                selfLine2DS.addAll(raysNotHit1.stream().map(a -> new java.awt.geom.Line2D.Double(a.getRay().getOrigin().getX(), a.getRay().getOrigin().getY(),
                        a.getRay().getEndPoint().getX(), a.getRay().getEndPoint().getY())).collect(Collectors.toSet()));
                stuckDto.setRaysNotHits(line2DS);
                stuckDto.setSelfRaysNotHits(selfLine2DS);
                stuckDto.setOpenPartCenter(Util.convertPoint(openPartCenter));
                stuckDto.setTargetValidRoads(Util.fetchIdValueFromElementIds(targetValidRoads.stream().map(CSURoad::getId).collect(Collectors.toSet())));
            }
            DebugHelper.VD_CLIENT.drawAsync(world.getSelfHuman().getID().getValue(), "StuckDtoLayer", stuckDto);
        }
        return moveToPoint(target);
    }

    private List<Point2D> getValidIntersections(Set<CSULineOfSightPerception.CsuRay> a, Set<CSULineOfSightPerception.CsuRay> b,
                                                Collection<CSURoad> targetValidRoads) {
        List<Point2D> intersections = Util.getSegmentIntersections(a, b);
        if (!intersections.isEmpty()) {
            filterIntersectionsNotInRoads(intersections, targetValidRoads);
        }
        return intersections;
    }

    private Collection<CSURoad> getSelfAndNearestRoad() {
        CSURoad csuRoad = world.getCsuRoad(world.getSelfPosition());
        Collection<CSURoad> result = new HashSet<>();
        result.add(csuRoad);
        result.add(world.getNearestNeighborRoad());
        return result;
    }

    private void filterIntersectionsNotInRoads(Collection<Point2D> points , Collection<CSURoad> roads) {
        Collection<Point2D> toRemove = new HashSet<>();
        for (Point2D point : points) {
            boolean contain = false;
            for (CSURoad road : roads) {
                Polygon polygon = road.getPolygon();
                if (polygon.contains(point.getX(), point.getY())) {
                    contain = true;
                    break;
                }
            }
            if (!contain) {
                toRemove.add(point);
            }
        }
        points.removeAll(toRemove);
    }

    private Action moveToPoint(Point2D target) {
        if (target == null) {
            return null;
        }
        List<EntityID> path = new ArrayList<>();
        path.add(world.getSelfPosition().getID());
        return new ActionMove(path, (int) target.getX(), (int) target.getY());
    }

    private TeamDooddCompositeConvexHull getConvexByBlockades(List<StandardEntity> blockades) {
        TeamDooddCompositeConvexHull convexHull = new TeamDooddCompositeConvexHull();
        for (StandardEntity entity : blockades) {
            Blockade blockade;
            if (entity instanceof Blockade) {
                blockade = (Blockade) entity;
                for (int i = 0; i < blockade.getApexes().length; i += 2) {
                    convexHull.addPoint(blockade.getApexes()[i], blockade.getApexes()[i + 1]);
                }
            }
        }
        return convexHull;
    }

    private List<EntityID> getSelfAndNeighborBlockades() {
        StandardEntity selfPosition = world.getSelfPosition();
        List<EntityID> blockadeIDs = new ArrayList<>();

        if (selfPosition instanceof Road) {
            Road road = (Road) selfPosition;
            for (StandardEntity next : world.getEntities(road.getNeighbours())) {
                if (next instanceof Road) {
                    Road neighbour = (Road) next;
                    if (neighbour.isBlockadesDefined()) {
                        blockadeIDs.addAll(neighbour.getBlockades());
                    }
                }
            }
            blockadeIDs.addAll(road.getBlockades());
            return blockadeIDs;
        }
        if(selfPosition instanceof  Building){
            Building building = (Building) selfPosition;
            for (StandardEntity next : world.getEntities(building.getNeighbours())) {
                if (next instanceof Road) {
                    Road neighbour = (Road) next;
                    if (neighbour.isBlockadesDefined()) {
                        blockadeIDs.addAll(neighbour.getBlockades());
                    }
                }
            }
            return blockadeIDs;
        }
        return null;
    }

    private EntityID getNearestBlockade() {
        List<EntityID> selfAndNeighborBlockades = getSelfAndNeighborBlockades();
        EntityID nearest = null;
        if (selfAndNeighborBlockades != null) {
            nearest = FbUtilities.getNearest(world, selfAndNeighborBlockades, world.getSelfHuman().getID());
        }
        return nearest;
    }

    private List<StandardEntity> getBlockadesInRange(EntityID entity, List<EntityID> entities, int range) {
        List<StandardEntity> inRangeBlockades = new ArrayList<>();
        if (entities == null || entities.isEmpty()) {
            return inRangeBlockades;
        }
        Blockade anchorBlockade = (Blockade) world.getEntity(entity);
        for (EntityID blockadeID : entities) {
            StandardEntity standardEntity = world.getEntity(blockadeID);
            if (!(standardEntity instanceof Blockade)) {
                continue;
            }
            Blockade blockade = (Blockade) standardEntity;
            double dist = Ruler.getDistance(Util.getPolygon(blockade.getApexes()), Util.getPolygon(anchorBlockade.getApexes()));
            if (dist < range) {
                inRangeBlockades.add(blockade);
            }
        }
        return inRangeBlockades;
    }

    private static class DistanceComparator implements Comparator<Point2D> {
        private Point2D reference;
        public DistanceComparator(Point2D reference) {
            this.reference = reference;
        }
        @Override
        public int compare(Point2D a, Point2D b) {
            int d1 = (int) Ruler.getDistance(reference, a);
            int d2 = (int) Ruler.getDistance(reference, b);
            return d1 - d2;
        }
    }
}
