package TeamDoodd_2021.module.complex.fb;

import TeamDoodd_2021.TeamDooddConstants;
import TeamDoodd_2021.debugger.DebugHelper;
import TeamDoodd_2021.module.complex.fb.FireBrigadeBuilding;
import TeamDoodd_2021.util.ambulancehelper.TeamDooddSelectorTargetByDis;
import adf.agent.communication.MessageManager;
import adf.agent.communication.standard.bundle.MessageUtil;
import adf.agent.communication.standard.bundle.StandardMessage;
import adf.agent.communication.standard.bundle.StandardMessagePriority;
import adf.agent.communication.standard.bundle.information.MessageAmbulanceTeam;
import adf.agent.communication.standard.bundle.information.MessageCivilian;
import adf.agent.communication.standard.bundle.information.MessageFireBrigade;
import adf.agent.communication.standard.bundle.information.MessagePoliceForce;
import adf.agent.develop.DevelopData;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.component.communication.CommunicationMessage;
import adf.component.module.AbstractModule;
import adf.component.module.algorithm.Clustering;
import adf.component.module.algorithm.PathPlanning;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;

import javax.swing.text.html.parser.Entity;
import java.io.Serializable;
import java.util.*;

public class CSUFBHumanSystem extends AbstractModule {

    private Clustering clustering;
    private PathPlanning pathPlanning;
    private Set<Civilian> allCivilian;

    public Map<EntityID, TeamDooddFBHuman> getAllHuman() {
        return allHuman;
    }

    private Map<EntityID, TeamDooddFBHuman> allHuman;
    private Set<Civilian> targetsCivilian;
    private Set<TeamDooddFBHuman> firstLevelCivilian;
    private Set<TeamDooddFBHuman> topLevelCivilian;
    private Set<TeamDooddFBHuman> secondLevelCivilian;
    private Set<TeamDooddFBHuman> targetHumans;
    private Set<TeamDooddFBHuman> removeHuman;
    private Set<Civilian> sendCivilian;
    private Set<Civilian> notNeedSendCivilian;
    public CSUFBHumanSystem(AgentInfo ai, WorldInfo wi, ScenarioInfo si, ModuleManager moduleManager, DevelopData developData, Clustering clustering, PathPlanning pathPlanning) {
        super(ai, wi, si, moduleManager, developData);
        this.allCivilian = new HashSet<>();
        this.allHuman = new HashMap<>();
        this.targetHumans = new HashSet<>();
        this.targetsCivilian = new HashSet<>();
        this.clustering = clustering;
        this.pathPlanning = pathPlanning;
        this.removeHuman = new HashSet<>();
        this.topLevelCivilian = new HashSet<>();
        this.firstLevelCivilian = new HashSet<>();
        this.secondLevelCivilian = new HashSet<>();
        this.sendCivilian = new HashSet<>();
        this.notNeedSendCivilian = new HashSet<>();
    }

    @Override
    public AbstractModule calc() {
        return null;
    }

    @Override
    public AbstractModule updateInfo(MessageManager messageManager) {
        if (this.getCountUpdateInfo() >= 2) {
            return this;
        }
        targetsCivilian.clear();
        targetHumans.clear();
        topLevelCivilian.clear();
        firstLevelCivilian.clear();
        secondLevelCivilian.clear();
        Set<EntityID> changed = this.agentInfo.getChanged().getChangedEntities();
        for (EntityID entityID : changed) {
            StandardEntity entity = this.worldInfo.getEntity(entityID);
            if (entity instanceof Human && !this.allHuman.containsKey(entityID)) {
                Human human = (Human) entity;
                if ((human.isBuriednessDefined() && human.getBuriedness() > 0) && (human.isHPDefined() && human.getHP() > 0)) {
                    TeamDooddFBHuman TeamDooddFBHuman = new TeamDooddFBHuman(human, agentInfo, worldInfo, scenarioInfo, pathPlanning, clustering);
                    this.allHuman.put(entityID, TeamDooddFBHuman);
                }
            }
            if (entity instanceof Civilian){
                Civilian civilian = (Civilian) entity;
                for (EntityID entityID1 : changed){
                    StandardEntity standardEntity = this.worldInfo.getEntity(entityID);
                    if (standardEntity instanceof AmbulanceTeam && ((AmbulanceTeam) standardEntity).getPosition().equals(civilian.getPosition())){
                        this.notNeedSendCivilian.add(civilian);
                    }
                }
                if (civilian.isBuriednessDefined()&&civilian.getBuriedness()==0&& civilian.isDamageDefined() && civilian.getDamage()>0&&civilian.isHPDefined()&&civilian.getHP()>0){
                    if (this.notNeedSendCivilian.contains(civilian)){
                        continue;
                    }
                    this.sendCivilian.add(civilian);
                }
            }
            if (entity instanceof AmbulanceTeam){
                for (Civilian civilian : this.sendCivilian){
                    if (this.notNeedSendCivilian.contains(civilian)){
                        continue;
                    }
                    if (civilian.isBuriednessDefined()&&civilian.getBuriedness()==0&& civilian.isDamageDefined() && civilian.getDamage()>0&&civilian.isHPDefined()&&civilian.getHP()>0){
                        messageManager.addMessage(new MessageCivilian(false, StandardMessagePriority.HIGH,civilian));
                    }
                }
                sendCivilian.clear();
            }
        }

        for (CommunicationMessage message : messageManager
                .getReceivedMessageList(MessageCivilian.class, MessageFireBrigade.class, MessagePoliceForce.class, MessageAmbulanceTeam.class)) {
            StandardEntity entity = null;
            entity = MessageUtil.reflectMessage(worldInfo,
                    (StandardMessage) message);
            if ((entity instanceof Human) && !this.allHuman.containsKey(entity.getID())) {
                Human human = (Human) entity;
                if ((human.isBuriednessDefined() && human.getBuriedness() > 0) && (human.isHPDefined() && human.getHP() > 0)) {
                    TeamDooddFBHuman TeamDooddFBHuman = new TeamDooddFBHuman(human, agentInfo, worldInfo, scenarioInfo, pathPlanning, clustering);
                    this.allHuman.put(entity.getID(), TeamDooddFBHuman);
                }
            }
        }
        //

        EntityID position = this.agentInfo.getPosition();
        for (TeamDooddFBHuman TeamDooddFBHuman : this.allHuman.values()) {
            Human human = TeamDooddFBHuman.getHuman();
            if ((human.isBuriednessDefined() && human.getBuriedness() <= 0) || (human.isHPDefined() && human.getHP() <= 0)) {
                removeHuman.add(TeamDooddFBHuman);
            } else if (human.getPosition().getValue() == position.getValue() && !changed.contains(human.getID())) {
                removeHuman.add(TeamDooddFBHuman);
            }
        }
        for (TeamDooddFBHuman TeamDooddFBHuman : removeHuman) {
            this.allHuman.remove(TeamDooddFBHuman.getHuman().getID());
        }
        int indexOfAgent = clustering.getClusterIndex(agentInfo.getID());
        for (TeamDooddFBHuman TeamDooddFBHuman : this.allHuman.values()) {
            TeamDooddFBHuman.updateInfo(messageManager);

            if (TeamDooddFBHuman.isReachable() && TeamDooddFBHuman.isCanRescue()){
                if ((agentInfo.getTime()<100&&!(TeamDooddFBHuman.getHuman() instanceof Civilian)) ||
                        ((TeamDooddFBHuman.getHuman() instanceof Civilian)&&!(TeamDooddFBHuman.getDeadTime()>200))){
                    if (TeamDooddFBHuman.getDistanceToMe()<50000&&TeamDooddFBHuman.getDeadTime()<60){
                        this.topLevelCivilian.add(TeamDooddFBHuman);
                    }
                    if (TeamDooddFBHuman.getDeadTime()<30&&TeamDooddFBHuman.getDistanceToMe()<240000){
                        this.firstLevelCivilian.add(TeamDooddFBHuman);
                    }else if(TeamDooddFBHuman.getDeadTime()<60&&TeamDooddFBHuman.getDistanceToMe()<120000){
                        this.secondLevelCivilian.add(TeamDooddFBHuman);
                    }
                }
                this.targetHumans.add(TeamDooddFBHuman);
            }
        }
        if (DebugHelper.DEBUG_MODE) {
            List<Integer> elements = new ArrayList<>();

            for (TeamDooddFBHuman TeamDooddFBHuman : this.allHuman.values()) {
                elements.add(TeamDooddFBHuman.getHuman().getID().getValue());
            }
            DebugHelper.VD_CLIENT.drawAsync(agentInfo.getID().getValue(), "ATPerceiveHumans", (Serializable) elements);
        }
        if (DebugHelper.DEBUG_MODE) {
            List<Integer> elements = new ArrayList<>();

            for (TeamDooddFBHuman TeamDooddFBHuman : this.targetHumans) {
                elements.add(TeamDooddFBHuman.getHuman().getID().getValue());
            }
            DebugHelper.VD_CLIENT.drawAsync(agentInfo.getID().getValue(), "ATTargetHumans", (Serializable) elements);
        }
        return this;
    }

    public EntityID getTargetByDistance() {
        EntityID target = null;
        TreeSet<TeamDooddFBHuman> objects3 = new TreeSet<>(
                ((o1, o2) -> Double.compare(o2.getValue(),o1.getValue())));
        objects3.addAll(this.topLevelCivilian);
        if (!objects3.isEmpty()){
            target = objects3.first().getHuman().getID();
            return target;
        }
        TreeSet<TeamDooddFBHuman> objects = new TreeSet<>(
                ((o1, o2) -> Double.compare(o2.getValue(),o1.getValue())));
        objects.addAll(this.firstLevelCivilian);
        if (!objects.isEmpty()){
            target = objects.first().getHuman().getID();
            return target;
        }
        TreeSet<TeamDooddFBHuman> objects1 = new TreeSet<>(
                ((o1, o2) -> Double.compare(o2.getValue(),o1.getValue())));
        objects1.addAll(this.secondLevelCivilian);
        if (!objects1.isEmpty()){
            target = objects1.first().getHuman().getID();
        }
        TreeSet<TeamDooddFBHuman> objects2 = new TreeSet<>(
                ((o1, o2) -> Double.compare(o2.getValue(),o1.getValue())));
        objects2.addAll(this.targetHumans);
        if (!objects2.isEmpty()){
            target = objects2.first().getHuman().getID();
        }
        return target;
    }


}

